(*
  writeatr - transfer ATR images from/to Atari floppy disks

  Copyright (c) 2002-2003 by Matthias Reichl <hias@horus.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

MODULE WriteATR;

IMPORT IO,AtariHD,Lib,FIO,Str;


FROM Storage IMPORT ALLOCATE,DEALLOCATE;

VAR DebugFlag : BOOLEAN;

TYPE AtrHeader = RECORD
         Header: CARDINAL;
         SizeLo : CARDINAL;
         SectorSize : CARDINAL;
         SizeHi : CARDINAL;
         Flags : BYTE;
         BadSector : CARDINAL;
         Unused : ARRAY [0..4] OF BYTE
     END;

(*
CONST SectorsPerTrack = 18;
      TracksPerDisk = 40;
      TotalSectors = SectorsPerTrack * TracksPerDisk;
*)

CONST MaxSectorsPerTrack = 100;

VAR buf:ARRAY[0..MaxSectorsPerTrack*256-1] OF BYTE;

PROCEDURE Hex2(b : SHORTCARD);
BEGIN
    IF (b<16) THEN
       IO.WrChar('0');
    END;
    IO.WrShtHex(b,0);
    IO.WrChar(' ');
END Hex2;

PROCEDURE ReadHeader(VAR file : FIO.File;
        name: ARRAY OF CHAR;
        VAR compact : BOOLEAN;
        VAR sectors : CARDINAL) : BOOLEAN ;
VAR len : CARDINAL;
    header : AtrHeader;
    bytes : LONGCARD;
    bps : CARDINAL;
BEGIN
    bps := AtariHD.GetBytesPerSector();

    len := FIO.RdBin(file, header, SIZE(header));
    IF (len # SIZE(header)) THEN
       IO.WrStr("error reading header of ATR-file "+'"');
       IO.WrStr(name);
       IO.WrStr('"');
       IO.WrLn;
       RETURN FALSE;
    END;
    IF DebugFlag THEN
        IO.WrStr("ATR-Header of ");
        IO.WrStr('"'); IO.WrStr(name); IO.WrStr('" :');
        IO.WrLn;
        Hex2(SHORTCARD(header.Header MOD 256));
        Hex2(SHORTCARD(header.Header DIV 256));
        Hex2(SHORTCARD(header.SizeLo MOD 256));
        Hex2(SHORTCARD(header.SizeLo DIV 256));
        Hex2(SHORTCARD(header.SectorSize MOD 256));
        Hex2(SHORTCARD(header.SectorSize DIV 256));
        Hex2(SHORTCARD(header.SizeHi MOD 256));
        Hex2(SHORTCARD(header.SizeHi DIV 256));
        Hex2(header.Flags);
        Hex2(SHORTCARD(header.BadSector MOD 256));
        Hex2(SHORTCARD(header.BadSector DIV 256));
        Hex2(header.Unused[0]);
        Hex2(header.Unused[1]);
        Hex2(header.Unused[2]);
        Hex2(header.Unused[3]);
        Hex2(header.Unused[4]);
        IO.WrLn;
        IO.WrStr("File size: ");
        IO.WrLngCard(FIO.Size(file),1);
        IO.WrLn;
    END;

    IF (header.Header # 0296H) THEN
       IO.WrStr('"'); IO.WrStr(name); IO.WrStr('"');
       IO.WrStr(" is not a valid ATR-file!");
       IO.WrLn;
       RETURN FALSE;
    END;
    IF (header.SectorSize <> 128) AND (header.SectorSize <> 256) THEN
       IO.WrStr('"'); IO.WrStr(name); IO.WrStr('"');
       IO.WrStr(": illegal sector size ");
       IO.WrCard(header.SectorSize,0);
       IO.WrLn;
       RETURN FALSE;
    END;
    IF (header.SectorSize # bps) THEN
       IO.WrStr('"'); IO.WrStr(name); IO.WrStr('"');
       IO.WrStr(": sector size of ATR file doesn't match selected format!");
       IO.WrLn;
       RETURN FALSE;
    END;
    bytes := (LONGCARD(header.SizeLo) + LONGCARD(header.SizeHi)*65536)*16;

    IF (bps = 256) THEN
        IF (bytes MOD LONGCARD(bps) = 128) THEN
           IF (bytes <= 384) THEN
               IO.WrStr("invalid size entry in ATR-file ");
               IO.WrStr('"'); IO.WrStr(name); IO.WrStr('"');
               IO.WrLn;
               RETURN FALSE;
           END;
           bytes := bytes + 384;
           compact := TRUE;
        ELSIF (bytes MOD LONGCARD(bps) # 0) THEN
           IO.WrStr("invalid size entry in ATR-file ");
           IO.WrStr('"'); IO.WrStr(name); IO.WrStr('"');
           IO.WrLn;
           RETURN FALSE;
        ELSE
           (* compact := FALSE; *)
           compact := TRUE;
        END;
    ELSE
        compact := TRUE;
        IF (bytes MOD LONGCARD(bps) # 0) THEN
           IO.WrStr("invalid size entry in ATR-file ");
           IO.WrStr('"'); IO.WrStr(name); IO.WrStr('"');
           IO.WrLn;
           RETURN FALSE;
        ELSE
           compact := TRUE;
        END;
    END;
    sectors := CARDINAL(bytes DIV LONGCARD(bps));
    IF DebugFlag THEN
        IO.WrStr("Sectors: ");
        IO.WrCard(sectors,1);
        IO.WrLn;
        IO.WrStr("Sector size: ");
        IO.WrCard(bps,0);
        IO.WrLn;
(*
        IO.WrStr("Compact: ");
        IO.WrBool(compact,0);
*)
        IO.WrLn;
        IO.WrLn;
    END;
    RETURN TRUE;
END ReadHeader;

PROCEDURE ReadAtrSector(
    VAR f: FIO.File; sec, maxSec : CARDINAL; compact : BOOLEAN)
 : BOOLEAN ;

VAR s,i,len : CARDINAL;
    pos : LONGCARD;
    bps : CARDINAL;
BEGIN
    s := (sec - 1) MOD AtariHD.GetSectorsPerTrack();

    bps := AtariHD.GetBytesPerSector();

    IF (sec <= maxSec) THEN
        IF (bps = 256) THEN
            IF compact THEN
                IF (sec <= 3) THEN
                    pos := LONGCARD(sec-1)*128;
                ELSE
                    pos := LONGCARD(sec-4)*256+384;
                END;
            ELSE
                pos := LONGCARD(sec-1)*256;
            END;
        ELSE
            pos := LONGCARD(sec-1)*LONGCARD(bps);
        END;

        FIO.Seek(f,SIZE(AtrHeader)+pos);

        IF (bps = 256) AND (sec <=3) AND compact THEN
            FOR i:=0 TO 127 DO
                buf[i+128+s*bps]:=0;
            END;
            len:=FIO.RdBin(f,buf[s*bps],128);
            IF (len # 128) THEN
                IO.WrLn;
                IO.WrStr("error reading ATR-file!"); IO.WrLn;
                RETURN FALSE;
            END;
        ELSE
            len:=FIO.RdBin(f,buf[s*bps],bps);
            IF (len # bps) THEN
                IO.WrLn;
                IO.WrStr("error reading ATR-file!"); IO.WrLn;
                RETURN FALSE;
            END;
        END;
        RETURN TRUE;
    ELSE
        FOR i:=0 TO bps-1 DO
            buf[i+s*bps]:=0;
        END;
        RETURN TRUE;
    END;
END ReadAtrSector;

PROCEDURE WriteImage( name : ARRAY OF CHAR; doVerify: BOOLEAN) ;

LABEL exit;

VAR f : FIO.File;
    compact : BOOLEAN;
    tracksPerDisk : CARDINAL;
    maxSectors : CARDINAL;
    secPerTrack : CARDINAL;

    headerOK : BOOLEAN;
    i, track : CARDINAL;

    diskSectors : CARDINAL;

BEGIN
    IF ~FIO.Exists(name) THEN
        IO.WrStr("file "+'"');
        IO.WrStr(name);
        IO.WrStr('" does not exist!');
        IO.WrLn;
        RETURN;
    END;

    f:=FIO.Open(name);
    IF ~FIO.OK THEN
        IO.WrStr("error opening file "+'"');
        IO.WrStr(name);
        IO.WrStr('"');
        IO.WrLn;
        RETURN;
    END;

    headerOK := ReadHeader(f, name, compact, maxSectors);

    IF (~headerOK) THEN
        GOTO exit;
    END;

    tracksPerDisk := AtariHD.GetTracksPerSide() *
                 AtariHD.GetSides();

    secPerTrack := AtariHD.GetSectorsPerTrack();

    IF DebugFlag THEN
        IO.WrStr("Going to write a total of ");
        IO.WrCard(tracksPerDisk,0);
        IO.WrStr(" tracks of ");
        IO.WrCard(AtariHD.GetSectorsPerTrack(),0);
        IO.WrStr(" sectors on ");
        IO.WrCard(AtariHD.GetSides(),0);
        IO.WrStr(" side(s)."); IO.WrLn;
        IO.WrLn;
    END;

    diskSectors := AtariHD.GetSectorsPerTrack()
                   * AtariHD.GetTracksPerSide()
                   * AtariHD.GetSides();

    IF (maxSectors < diskSectors) THEN
       IO.WrStr("Warning: ATR image is smaller than selected disk."); IO.WrLn;
       IO.WrStr("         (ATR: ");
       IO.WrCard(maxSectors,0);
       IO.WrStr(" sectors, disk: ");
       IO.WrCard(diskSectors,0);
       IO.WrStr(" sectors)"); IO.WrLn;
       IO.WrStr("         Padding remaining sectors with zeros."); IO.WrLn;
       IO.WrLn;
    END;

    IF (maxSectors > diskSectors) THEN
       IO.WrStr("Warning: ATR image is larger than selected disk."); IO.WrLn;
       IO.WrStr("         (ATR: ");
       IO.WrCard(maxSectors,0);
       IO.WrStr(" sectors, disk: ");
       IO.WrCard(diskSectors,0);
       IO.WrStr(" sectors)"); IO.WrLn;
       IO.WrStr("         !!! Will only write the first ");
       IO.WrCard(diskSectors,0);
       IO.WrStr(" sectors to disk !!!"); IO.WrLn;
       IO.WrLn;
    END;

    IO.WrStr("Writing with");
    IF NOT doVerify THEN
       IO.WrStr("out");
    END;
    IO.WrStr(" verify: ");
    IO.WrChar('"'); IO.WrStr(name); IO.WrChar('"');
    IO.WrLn;
    IO.WrCard(0,3); IO.WrStr("%");

    FOR track := 0 TO tracksPerDisk-1 DO

        AtariHD.FormatTrack(track);
        IF AtariHD.ok = FALSE THEN
            IO.WrLn;
            IO.WrStr("error formatting track - aborting!");
            IO.WrLn;
            GOTO exit;
        END;
        FOR i:=0 TO secPerTrack-1 DO
            IF (ReadAtrSector(f, track*secPerTrack+i+1, maxSectors, compact) = FALSE) THEN
                GOTO exit;
            END;
        END;

        AtariHD.WriteTrack(track,buf);
        IF ~AtariHD.ok THEN
            IO.WrLn;
            IO.WrStr("error writing track!"); IO.WrLn;
            GOTO exit;
        END;

        IF (doVerify) THEN
            AtariHD.VerifyTrack(track,buf);
            IF ~AtariHD.ok THEN
                IO.WrLn;
                IO.WrStr("error verifying track - aborting!"); IO.WrLn;
                GOTO exit;
            END;
        END;

        IO.WrCharRep(CHAR(8),4);
        IO.WrCard((100*(track+1)) DIV tracksPerDisk,3); IO.WrStr("%");
    END;
    IO.WrLn;
    IO.WrLn;
    IO.WrStr("successfully wrote disk!");
    IO.WrLn;

exit:
    FIO.Close(f);
END WriteImage;

PROCEDURE ReadImage( name : ARRAY OF CHAR);

LABEL exit, exitWrite;

VAR f : FIO.File;
    secPerTrack : CARDINAL;
    tracksPerDisk : CARDINAL;
    bytesPerDisk : LONGCARD;

    headerOK : BOOLEAN;
    i, track : CARDINAL;

    myHeader : AtrHeader;

    bps : CARDINAL;
BEGIN

    bps := AtariHD.GetBytesPerSector();

    f:=FIO.Create(name);
    IF ~FIO.OK THEN
        IO.WrStr("error creating file "+'"');
        IO.WrStr(name);
        IO.WrStr('"');
        IO.WrLn;
        RETURN;
    END;

    secPerTrack := AtariHD.GetSectorsPerTrack();
    tracksPerDisk := AtariHD.GetTracksPerSide() *
                     AtariHD.GetSides();

    bytesPerDisk := LONGCARD(secPerTrack)*LONGCARD(tracksPerDisk)*LONGCARD(bps);
    IF (bps = 256) THEN
        bytesPerDisk := bytesPerDisk-384;
    END;

    myHeader.Header := 0296H;

    myHeader.SizeLo := CARDINAL((bytesPerDisk DIV 16) MOD 65536);
    myHeader.SizeHi := CARDINAL((bytesPerDisk DIV 16) DIV 65536);

    myHeader.SectorSize := bps;
    myHeader.Flags := 00H;
    myHeader.BadSector := 0000H;
    myHeader.Unused[0] := 0H;
    myHeader.Unused[1] := 0H;
    myHeader.Unused[2] := 0H;
    myHeader.Unused[3] := 0H;
    myHeader.Unused[4] := 0H;

    FIO.WrBin(f, myHeader, SIZE(myHeader));
    IF (~FIO.OK) THEN
        GOTO exitWrite;
    END;


    IF DebugFlag THEN
        IO.WrStr("Going to read a total of ");
        IO.WrCard(tracksPerDisk,0);
        IO.WrStr(" tracks of ");
        IO.WrCard(AtariHD.GetSectorsPerTrack(),0);
        IO.WrStr(" sectors on ");
        IO.WrCard(AtariHD.GetSides(),0);
        IO.WrStr(" sides."); IO.WrLn;
        IO.WrLn;
    END;


    IO.WrStr("Reading: ");
    IO.WrChar('"'); IO.WrStr(name); IO.WrChar('"');
    IO.WrLn;
    IO.WrCard(0,3); IO.WrStr("%");

    FOR track := 0 TO tracksPerDisk-1 DO

        AtariHD.ReadTrack(track, buf);
        IF (~AtariHD.ok) THEN
            IO.WrLn;
            IO.WrStr("error reading track - aborting!");
            IO.WrLn;
            GOTO exit;
        END;

        IF (track = 0) AND (bps=256) THEN
            FIO.WrBin(f, buf, 128);
            IF (~FIO.OK) THEN
                GOTO exitWrite;
            END;
            FIO.WrBin(f, buf[256], 128);
            IF (~FIO.OK) THEN
                GOTO exitWrite;
            END;
            FIO.WrBin(f, buf[512], 128);
            IF (~FIO.OK) THEN
                GOTO exitWrite;
            END;
            FIO.WrBin(f, buf[768], (secPerTrack-3) * bps);
            IF (~FIO.OK) THEN
                GOTO exitWrite;
            END;
        ELSE
            FIO.WrBin(f, buf, secPerTrack * bps);
            IF (~FIO.OK) THEN
                GOTO exitWrite;
            END;
        END;
        IO.WrCharRep(CHAR(8),4);
        IO.WrCard((100*(track+1)) DIV tracksPerDisk,3); IO.WrStr("%");
    END;

    FIO.Close(f);
    IO.WrLn;
    IO.WrLn;
    IO.WrStr("successfully created image!");
    IO.WrLn;
    RETURN;

exitWrite:

    IO.WrLn;
    IO.WrStr("error writing ATR-file!");
    IO.WrLn;

exit:
    FIO.Close(f);
    FIO.Erase(name);
END ReadImage;


VAR name,opt:ARRAY[0..255] OF CHAR;
    drive:CARDINAL;
    gotFile : BOOLEAN;
    i,j: CARDINAL;

    LABEL usage, usage2, invalidOption;

TYPE Mode = ( Write, WriteAndVerify, Read );

VAR currentMode : Mode;

TYPE AtariFloppyType = ( FloppyDD, FloppyDSDD, FloppyXF551DSDD );
VAR currentFloppyType : AtariHD.FloppyType;

VAR
    tmpBool : BOOLEAN;
    tmpLongInt : LONGINT;
    tmpString : ARRAY[0..255] OF CHAR;

    debugLevel : CARDINAL;

BEGIN
    IO.WrStr("WriteAtr V0.92b"); IO.WrLn;
    IO.WrStr("(c) 2003 by HiassofT - Matthias Reichl <hias@horus.com>"); IO.WrLn;
    IO.WrLn;
    j:=Lib.ParamCount();

    IF j<1 THEN
       GOTO usage2;
usage:
       IO.WrLn;
usage2:
       IO.WrStr('Usage: writeatr [-b] [-d] [-fX] [-n] [-r] filename.atr'); IO.WrLn;
       IO.WrLn;
       IO.WrStr("options:"); IO.WrLn;
       IO.WrStr("    -b : use drive B"); IO.WrLn;
       IO.WrStr("    -fX : set atari floppy drive type (X=0..8)"); IO.WrLn;
       IO.WrStr('          0=180k/5.25" (this is the default)'); IO.WrLn;
       IO.WrStr('          1=360k/5.25"        2=720k/3.5"        3=1.44M/3.5"'); IO.WrLn;
       IO.WrStr('          4=360k/5.25"/head   5=720k/3.5"/head   6=1.44M/3.5"/head'); IO.WrLn;
       IO.WrStr('          7=360k/5.25"/XF551  8=720k/3.5"/XF551'); IO.WrLn;
       IO.WrStr('          9=130k/5.25" (1050 enhanced density) EXPERIMENTAL!'); IO.WrLn;
       IO.WrStr("    -r : read disk (create ATR-image of disk)"); IO.WrLn;
       IO.WrStr("    -d : increase debug level by 1"); IO.WrLn;
       IO.WrStr("    -n : no verify when writing (DANGEROUS!)"); IO.WrLn;
       HALT
    END;

    drive := 0;
    gotFile := FALSE;
    DebugFlag := FALSE;
    debugLevel := 0;
    currentMode := WriteAndVerify;
    currentFloppyType := AtariHD.FloppyDD;

    FOR i := 1 TO j DO
        Lib.ParamStr(opt,i);
        IF opt[0]='-' THEN
            IF Str.Length(opt) < 2 THEN
                GOTO invalidOption;
            END;
            CASE CAP(opt[1]) OF
            | 'B' :
                IF Str.Length(opt) # 2 THEN
                    GOTO invalidOption;
                ELSE
                    drive := 1;
                END;
            | 'D' :
                IF Str.Length(opt) # 2 THEN
                    GOTO invalidOption;
                ELSE
                    debugLevel := debugLevel + 1;
                END;
            | 'R' :
                IF Str.Length(opt) # 2 THEN
                    GOTO invalidOption;
                ELSE
                    currentMode := Read;
                END;
            | 'N' :
                IF Str.Length(opt) # 2 THEN
                    GOTO invalidOption;
                ELSE
                    IF (currentMode = WriteAndVerify) THEN
                        currentMode := Write;
                    END;
                END;
            | 'F' :
                IF (Str.Length(opt) <= 2) THEN
                    GOTO invalidOption;
                END;
                Str.Slice(tmpString,opt, 2, Str.Length(opt)-2);
                tmpLongInt := Str.StrToInt(tmpString,10,tmpBool);
                IF (~tmpBool) OR (tmpLongInt < 0) OR (tmpLongInt > 9) THEN
                    IO.WrStr("invalid drive type "+'"');
                    IO.WrStr(tmpString);
                    IO.WrStr('"');
                    IO.WrLn;
                    GOTO usage;
                END;
                currentFloppyType := AtariHD.FloppyType(tmpLongInt);
            ELSE
invalidOption:
                IO.WrStr("invalid option "+'"');
                IO.WrStr(opt);
                IO.WrStr('"');
                IO.WrLn;
                GOTO usage;
            END;
        ELSE
            IF (gotFile) THEN
                GOTO usage;
            END;
            gotFile := TRUE;
            Str.Copy(name,opt);
        END;
    END;

    IF ~gotFile THEN
        GOTO usage;
    END;

    IF (debugLevel >= ORD(AtariHD.DebugLevel_High)) THEN
       debugLevel := ORD(AtariHD.DebugLevel_High);
    END;
    IF (debugLevel >= 1) THEN
       DebugFlag := TRUE;
    END;

    AtariHD.InitAtariHD;

    AtariHD.SetDebugLevel(AtariHD.DebugLevelType(debugLevel));
    AtariHD.SetInteractive(TRUE);

    IF drive=0 THEN
       AtariHD.SetDriveA;
    ELSE
       AtariHD.SetDriveB;
    END;

    IO.WrStr("Selected Atari disk drive type: ");
    CASE currentFloppyType OF
    | AtariHD.FloppyDD : IO.WrStr('180k SS/DD 5.25"');
    | AtariHD.FloppyDSDD : IO.WrStr('360k DS/DD 5.25"');
    | AtariHD.Floppy720_35 : IO.WrStr('720k DS/DD 3.5"');
    | AtariHD.Floppy1440_35 : IO.WrStr('1.44M DS/HD 3.5"');
    | AtariHD.FloppyDSDD_HeadFirst : IO.WrStr('360k DS/DD 5.25" head first');
    | AtariHD.Floppy720_35_HeadFirst : IO.WrStr('720k DS/DD 3.5" head first');
    | AtariHD.Floppy1440_35_HeadFirst : IO.WrStr('1.44M DS/HD 3.5" head first');
    | AtariHD.FloppyDSDD_XF551 : IO.WrStr('360k DS/DD 5.25" XF551');
    | AtariHD.Floppy720_35_XF551 : IO.WrStr('720k DS/DD 3.5" XF551');
    | AtariHD.FloppyED : IO.WrStr('130k SS/SD 5.25" (1050 enhanced density)');
      IO.WrLn;
      IO.WrStr("  **************************************************************************");
      IO.WrLn;
      IO.WrStr("  *            WARNING: using enhanced density is EXPERIMENTAL             *");
      IO.WrLn;
      IO.WrStr("  * reading disks and writing with verify will fail with most controllers, *");
      IO.WrLn;
      IO.WrStr("  *                  writing without verify _might_ work                   *");
      IO.WrLn;
      IO.WrStr("  **************************************************************************");
    ELSE
        IO.WrStr("unknown - aborting!");
        HALT;
    END;
    IO.WrLn;
    IO.WrLn;

    IF (~AtariHD.CheckDriveCompatibility(currentFloppyType)) THEN
       IO.WrStr("The selected atari disk drive type is not compatible with"); IO.WrLn;
       IO.WrStr("your floppy drive!"); IO.WrLn;
       HALT;
    END;

    AtariHD.InitFloppy(currentFloppyType);

    IF (debugLevel = ORD(AtariHD.DebugLevel_High)) THEN
       AtariHD.PrintDebugInfo();
       IO.WrLn;
    END;

    CASE currentMode OF
    | Write :
        WriteImage(name, FALSE);
    | WriteAndVerify:
        WriteImage(name, TRUE);
    | Read:
        ReadImage(name);
    END;

END WriteATR.
